function [c,G,g,H,h,lb,ub] = DWempcSetupLP(A,B,C,Upsilon,x0,p,w,r,q,...
    u_min,u_max,Du_min,Du_max,y_min,y_max,yT_min,yT_max,gamma_max,...
    rho_max,eta_max,u_m1)
% setup DW block angular problem from MPC data

%% ========================================================================
% BLOCK-DIAGNOAL LP SETUP
% =========================================================================
M = length(A);
N = length(yT_max{1});
nx = cell(M,1);
nu = cell(M,1);
ny = cell(M,1);
for j = 1:M
    [nx{j},nu{j}] = size(B{j});
    ny{j} = size(C{j},1);
end

%% ========================================================================
% MPC MATRICES - LOCAL
% =========================================================================
Gamma = cell(M,1);
H = cell(M,1);
for j = 1:M
    [Gamma{j},H{j}] = MPCbuildGamma(A{j},B{j},C{j},N);
end

Phi = cell(M,1);
for j = 1:M
    Phi{j} = zeros(N*ny{j},nx{j});
    k1 = 1;
    for k = 1:N
        Phi{j}(k1:k1+ny{j}-1,:) = C{j}*A{j}^k;
        k1 = k1 + ny{j};
    end
end

%% ========================================================================
% MPC MATRICES - LINKING
% =========================================================================
Gamma_tilde = cell(M,1);
H_tilde = cell(M,1);
for j = 1:M
    [Gamma_tilde{j},H_tilde{j}] = MPCbuildGamma(A{j},B{j},...
        Upsilon{j}*C{j},N);
end

Phi_tilde = cell(M,1);
for j = 1:M
    Phi_tilde{j} = zeros(N*1,nx{j});
    k1 = 1;
    for k = 1:N
        Phi_tilde{j}(k1:k1+1-1,:) = Upsilon{j}*C{j}*A{j}^k;
        k1 = k1 + 1;
    end
end

%% ========================================================================
% BUILD LP BLOCKS
% =========================================================================
c = cell(M,1);

for j = 1:M
    c{j} = [expand(p{j}); expand(r{j}); expand(w{j})];
end
c{M+1,1} = expand(q{1});

Lambda = cell(M,1);
for j = 1:M
    Nb = N*nu{j};
    dgN = (1:Nb:Nb*Nb) + (0:Nb-1);
    dgL = (nu{j}:Nb:Nb*Nb-nu{j}*Nb) + (1:Nb-nu{j});
    Lambda{j} = zeros(Nb,Nb);
    Lambda{j}(dgN) = 1; Lambda{j}(dgL) = -1;
end

I0 = cell(M,1);
for j = 1:M
    I0{j} = zeros(nu{j}*N,nu{j});
    I0{j}(1:nu{j},1:nu{j}) = eye(nu{j});
end

G = cell(M+1,1);
g = cell(M+1,1);
lb = cell(M+1,1);
ub = cell(M+1,1);
h = cell(1,1);
H = cell(M+1,1);
for j = 1:M
    G{j} = [...
        Lambda{j} sparse(N*nu{j},N*ny{j}) sparse(N*nu{j},N*nu{j});
        -Lambda{j} sparse(N*nu{j},N*ny{j}) sparse(N*nu{j},N*nu{j});
        -Gamma{j} -speye(N*ny{j}) sparse(N*ny{j},N*nu{j});
        Gamma{j} -speye(N*ny{j}) sparse(N*ny{j},N*nu{j});
        -Lambda{j}  sparse(N*nu{j},N*ny{j}) -speye(N*nu{j})
        Lambda{j}  sparse(N*nu{j},N*ny{j}) -speye(N*nu{j})
        ];
    
    lb{j} = [expand(u_min{j}); zeros(ny{j}*N,1); zeros(nu{j}*N,1)];
    ub{j} = [expand(u_max{j}); expand(gamma_max{j}); expand(eta_max{j})];
    
    g{j} = [-expand([Du_min{j}]) - I0{j}*u_m1{j};
            expand(Du_max{j}) + I0{j}*u_m1{j}
            -expand([y_min{j}]) + Phi{j}*x0{j};
            expand(y_max{j}) - Phi{j}*x0{j};
            -I0{j}*u_m1{j}
            I0{j}*u_m1{j}
        ];
    
    H{j} = [-Gamma_tilde{j} sparse(N,N*ny{j}) sparse(N,N*nu{j})
            Gamma_tilde{j} sparse(N,N*ny{j}) sparse(N,N*nu{j})];
end
h{1} = [-expand(yT_min{1}) + [Phi_tilde{1:end}]*expand([x0{1:end}]);
        expand(yT_max{1}) - [Phi_tilde{1:end}]*expand([x0{1:end}])];
H{M+1,1} = [-eye(N); -eye(N)];
lb{M+1} = zeros(N,1);
ub{M+1} = expand(rho_max{1});

G{M+1,1} = zeros(N);
g{M+1,1} = zeros(N,1);

%% ========================================================================
%  REMOVE INF VALUES
% =========================================================================
L = inf; % for some solvers L may need to be replaced with a large value
for j = 1:M
   idx = isinf(abs(g{j}));
   G{j}(idx,:) = [];
   g{j}(idx,:) = [];
end
idx = isinf(abs(h{1}));
h{1}(idx,:) = [];
for j = 1:M
   H{j}(idx,:) = [];
end
for j = 1:M
    idx = isinf(c{j});
    c{j}(idx) = L;
    idx = isinf(-c{j});
    c{j}(idx) = -L;
end


function [Gamma,H] = MPCbuildGamma(A,B,C,N)
%% ------------------------------------------------------------------------
% Gamma Matrix
[ny,~] = size(C);
nu = size(B,2);
Gamma = zeros(N*ny,N*nu);

%% ------------------------------------------------------------------------
% Impulse Response
H = zeros(ny,nu,N);
for k = 1:N
    H(:,:,k) = C*A^(k-1)*B;
end

%% ------------------------------------------------------------------------
% Build Block Column 1
T = C;
k1 = 1;
k2 = ny;
for k=1:N
    Gamma(k1:k2,1:nu) = T*B;
    k1 = k1+ny;
    k2 = k2+ny;
    T = T*A;
end

%% ------------------------------------------------------------------------
% Build Block Column 2-N By Copy
k1 = ny+1;
k2 = nu+1;
k3 = k2+nu-1;
k4 = N*ny - ny;
for k=2:N
    Gamma(k1:end,k2:k3) = Gamma(1:k4,1:nu);
    k1 = k1+ny;
    k2 = k2+nu;
    k3 = k3+nu;
    k4 = k4-ny;
end


function x = expand(M,varargin)
% expands a vector/matrix
if(~isempty(varargin))
    transpose = varargin{:};
else 
    transpose = 0;
end

if(transpose)
    M = M';
end

if(isvector(M))
    x = M(:);
else
    Mv = M(:);
    x = Mv(:);
end


